from __future__ import annotations

import sys
from dataclasses import field, replace
from typing import Counter

import pytest

from ovld import recurse
from ovld.codegen import Code, Lambda, code_generator
from ovld.core import ovld
from ovld.dependent import Regexp
from ovld.medley import (
    BuildOvld,
    ChainAll,
    CodegenParameter,
    KeepLast,
    Medley,
    ReduceAll,
    RunAll,
    meld,
    use_combiner,
)

# Skip all tests if Python version is less than 3.10
pytestmark = pytest.mark.skipif(
    sys.version_info < (3, 10), reason="These tests require Python 3.10 or higher"
)


class Zero(Medley):
    def do(self, x: object):
        return 0


class Apple(Medley):
    worms: int = 1

    def do(self, x: int):
        return f"{x * self.worms} worms"


class Banana(Medley):
    rings: int

    def do(self, x: str):
        return " ".join([f"{x}!"] * self.rings)

    def do(self, x: object):
        return "~fallback~"


class CherryBomb(Medley):
    red: CodegenParameter[bool]

    @code_generator
    def do(cls, x: str):
        return Lambda(Code("$transform($x)", transform=str.upper if cls.red else str.lower))


class Dongle(Medley):
    def do(self, xs: list):
        return [recurse(x) for x in xs]


class DarkApple(Medley):
    worms: str = "worms"

    def do(self, x: int):
        return f"{x} {self.worms}"


def test_simple_mixin():
    a = Apple(worms=5)
    assert a.do(10) == "50 worms"
    with pytest.raises(TypeError, match="No method in <Ovld"):
        a.do("wat")


def test_simple_mixin_2():
    b = Banana(rings=3)
    assert b.do("ring") == "ring! ring! ring!"
    assert b.do(10) == "~fallback~"


def test_codegen():
    bt = CherryBomb(red=True)
    bf = CherryBomb(red=False)

    assert bt.do("Wow") == "WOW"
    assert bf.do("Wow") == "wow"


def test_meld():
    ab = meld([Apple(5), Banana(3)])
    assert ab.worms == 5
    assert ab.rings == 3
    assert ab.do(10) == "50 worms"
    assert ab.do("ring") == "ring! ring! ring!"


def test_recurse():
    ab = meld([Apple(5), Banana(3), Dongle()])
    assert ab.do([10, "ring"]) == ["50 worms", "ring! ring! ring!"]


def test_recurse_codegen():
    ab = meld([Apple(5), CherryBomb(True), Dongle()])
    assert ab.do([10, "ring"]) == ["50 worms", "RING"]


def test_meld_operator():
    ab = Apple(5) + Banana(3) + Dongle()
    assert ab.do([10, "ring"]) == ["50 worms", "ring! ring! ring!"]


def test_codegen_reuse():
    gens = Counter()

    class One(Medley):
        flag: CodegenParameter[bool]
        default: object

        @code_generator
        def do(cls, x: str):
            gens[One] += 1
            method = str.upper if cls.flag else str.lower
            return Lambda(Code("$method($x)", method=method))

        def do(self, x: object):
            return self.default

    class Two(Medley):
        factor: CodegenParameter[int]

        @code_generator
        def do(cls, x: int):
            gens[Two] += 1
            return Lambda(Code("$x * $factor", factor=cls.factor))

    obby = object()

    ot1 = meld([One(True, 51), Two(3)])
    ot2 = meld([One(True, "foo"), Two(3)])
    ot3 = meld([One(False, 51), Two(10)])

    assert type(ot1) is type(ot2)
    assert type(ot1) is not type(ot3)

    assert gens[One] == 0
    assert gens[Two] == 0

    assert ot1.do("Hello") == "HELLO"
    assert ot1.do("boop") == "BOOP"
    assert ot1.do(18) == 54
    assert ot1.do(obby) == 51

    assert gens[One] == 1
    assert gens[Two] == 1

    assert ot2.do("Hello") == "HELLO"
    assert ot2.do(18) == 54
    assert ot2.do(obby) == "foo"

    assert gens[One] == 1
    assert gens[Two] == 1

    assert ot3.do("Hello") == "hello"
    assert ot3.do(18) == 180
    assert ot3.do(obby) == 51

    assert gens[One] == 2
    assert gens[Two] == 2


def test_conflict():
    with pytest.raises(TypeError):
        meld([Apple(3), DarkApple(3)])


def test_meld_classes():
    Abba = Apple + Banana
    ab = Abba(worms=5, rings=3)
    assert ab.worms == 5
    assert ab.rings == 3
    assert ab.do(10) == "50 worms"
    assert ab.do("ring") == "ring! ring! ring!"


def test_meld_inplace():
    class One(Medley):
        x: int

        def do(self, x: int):
            return x * self.x

        def do(self, x: object):
            return "fallback"

    class Two(Medley):
        y: str = field(default="?")

        def do(self, x: str):
            return x + self.y

    before = One(x=5)
    assert before.do(10) == 50
    assert before.do("wow") == "fallback"

    One += Two

    # Existing instance should work with the new behavior and the defaults
    assert before.do("wow") == "wow?"

    # New instances can be created with the new fields
    onus = One(x=4, y="!!!!!!")
    assert onus.do(10) == 40
    assert onus.do("wow") == "wow!!!!!!"

    class Three(Medley):
        z: str  # <= should have a default value so that existing code doesn't break

        def do(self, x: str):
            return self.z

    with pytest.raises(TypeError, match="Dataclass field 'z' must have a default value"):
        One += Three

    # Unchanged because the melding failed
    assert onus.do("wow") == "wow!!!!!!"


def test_meld_inplace_keep_pointers():
    class One(Medley):
        x: int

        def do(self, x: int):
            return x * self.x

        def do(self, x: object):
            return "fallback"

    class Two(Medley):
        y: str = field(default="?")

        def do(self, x: str):
            return x + self.y

    before = One(x=5)

    saved_do = before.do
    assert saved_do(10) == 50
    assert saved_do("wow") == "fallback"

    assert saved_do.__ovld__ is before.do.__ovld__
    One += Two
    assert saved_do.__ovld__ is before.do.__ovld__

    # Existing method should work with the new behavior and the defaults
    assert saved_do(10) == 50
    assert saved_do("wow") == "wow?"


def test_meld_inplace_propagates_to_subclasses():
    class One(Medley):
        x: int

        def do(self, x: int):
            return x * self.x

        def do(self, x: object):
            return "fallback"

    class Two(Medley):
        y: str = field(default="?")

        def do(self, x: str):
            return x + self.y

    class Three(Medley):
        def do(self, x: float):
            return -x

    one3 = (One + Three)(x=7)

    assert one3.do("wow") == "fallback"
    One += Two
    assert one3.do("wow") == "wow?"
    assert one3.do(5.5) == -5.5


def test_post_init():
    class One(Medley):
        x: int

        def __post_init__(self):
            self.xx = self.x * self.x

        def do(self, x: int):
            return x * self.xx

    class Two(Medley):
        y: int

        def __post_init__(self):
            self.yy = self.y + self.y

        def do(self, x: str):
            return x + self.yy

    ot = One(3) + Two("v")
    assert ot.xx == 9
    assert ot.yy == "vv"
    assert ot.do(10) == 90
    assert ot.do("z") == "zvv"

    otc = One + Two
    ot2 = otc(x=3, y="v")
    assert ot2.xx == 9
    assert ot2.yy == "vv"
    assert ot2.do(10) == 90
    assert ot2.do("z") == "zvv"


def test_medley_isinstance():
    assert issubclass(Apple + Banana, Apple)
    assert isinstance(Apple(10) + Banana(7), Banana)


def test_medley_name():
    assert (Apple + Banana).__name__ == "Apple+Banana"


def test_medley_replace():
    ab = Apple(10) + Banana(7)
    assert ab.do(8) == "80 worms"

    ab2 = replace(ab, worms=100)
    assert ab2.do(8) == "800 worms"


def test_medley_replace_codegen():
    cb = CherryBomb(red=True)
    assert cb.red
    assert cb.do("Hello") == "HELLO"
    cb2 = replace(cb, red=False)
    assert not cb2.red
    assert cb2.do("Hello") == "hello"


def test_add_same_type():
    aa = Apple(10) + Apple(20)
    assert aa.do(8) == "160 worms"


def test_subtract():
    a = Apple(10)
    b = Banana(20)
    ab = a + b
    assert ab.worms == 10
    assert ab.rings == 20

    a2 = ab - Banana
    assert a2.worms == 10
    assert isinstance(a2, Apple)
    assert not isinstance(a2, Banana)
    assert not hasattr(a2, "rings")

    b2 = ab - Apple
    assert b2.rings == 20
    assert not isinstance(b2, Apple)
    assert isinstance(b2, Banana)
    assert not hasattr(b2, "worms")

    b3 = ab - Apple - Banana
    assert not hasattr(b3, "rings")
    assert not hasattr(b3, "worms")

    with pytest.raises(TypeError, match="unexpected keyword argument 'worms'"):
        ((Apple + Banana) - Apple)(worms=10, rings=20)

    with pytest.raises(TypeError, match="unexpected keyword argument 'rings'"):
        ((Apple + Banana) - Banana)(worms=10, rings=20)


def test_init_not_allowed():
    with pytest.raises(Exception, match="Do not define __init__"):

        class Grapes(Medley):
            def __init__(self, z: int, w=0):
                self.z = z
                self.w = w

            def do(self, x: int):
                return x + self.z + self.w


def test_inheritance():
    class Flapjack(Medley):
        def do(self, x: int):
            return x + 1

    class Cake(Flapjack):
        def do(self, x: str):
            return x + "s"

    assert Cake().do("egg") == "eggs"
    assert Cake().do(6) == 7


def test_priority():
    class Habanero(Medley):
        def do(self, x: int):
            return "A"

        @ovld(priority=1000)
        def do(self, x: int):
            return "B"

        def do(self, x: int):
            return "C"

    assert Habanero().do(3) == "B"


def test_reduce_all():
    class Iguana(Medley):
        @use_combiner(ReduceAll)
        def increment(self, x):
            return x + 1

        def increment(self, x):
            return x + 2

        def increment(self, x):
            return x + 7

    assert Iguana().increment(3) == 13


def test_chain_all():
    class Jellyfish(Medley):
        x: int = 0

        @use_combiner(ChainAll)
        def swim(self):
            self.x += 1
            return self

        def swim(self):
            self.x *= 2
            return self

        def swim(self):
            self.x += 3
            return self

    j = Jellyfish()
    j.swim()
    assert j.x == 5  # ((0 + 1) * 2) + 3


def test_class_cache():
    ab1 = Apple + Banana
    ab2 = Apple + Banana
    assert ab1 is ab2

    # Same overall order
    abc1 = (Apple + Banana) + CherryBomb
    abc2 = Apple + (Banana + CherryBomb)
    assert abc1 is abc2

    # Different order
    abc3 = Banana + (CherryBomb + Apple)
    assert abc1 is not abc3


def test_adding_superclasses_is_idempotent():
    class Plantain(Banana):
        yellow: bool = True

    assert (Banana + Plantain) is Plantain


def test_issubclass():
    class Plantain(Banana):
        yellow: bool = True

    assert issubclass(Apple + Plantain, Banana)
    ABC = Apple + Banana + CherryBomb
    assert issubclass(ABC, Apple + Banana)
    assert issubclass(ABC, Apple + CherryBomb)
    assert issubclass(ABC, Banana + CherryBomb)


def test_incompatible_combiners():
    class Kangaroo(Medley):
        jump = KeepLast()

    class Frog(Medley):
        jump = BuildOvld()

    with pytest.raises(TypeError):
        Kangaroo + Frog


def test_refer_to_defined_vars():
    class Llama(Medley):
        def do(self, x: int):
            return x * x

        spit = do

        def do(self, x: str):
            return x + "s"

    ll = Llama()
    assert ll.spit(10) == 100
    assert ll.spit("neat") == "neats"


def test_absent():
    class Lux(Medley):
        one = KeepLast()
        two = RunAll()
        three = BuildOvld()

    lx = Lux()

    with pytest.raises(AttributeError, match="no attribute 'one'"):
        lx.one
    with pytest.raises(AttributeError, match="no attribute 'two'"):
        lx.two
    with pytest.raises(AttributeError, match="no attribute 'three'"):
        lx.three


def test_default_factory_inheritance():
    class Lost(Medley):
        xs: list = field(default_factory=list)

    class Lust(Lost):
        pass

    lu = Lust()
    assert lu.xs == []


class Monkey(Medley):
    regex: CodegenParameter[str] = "monk[iey]+"

    def hi(self, x: Regexp[regex]):
        return "hello!"

    def hi(self, x: str):
        return "goodbye!"


def test_configure_dependent():
    m1 = Monkey()
    m2 = Monkey(regex="go+rilla")

    assert m1.hi("monki") == "hello!"
    assert m1.hi("monkeeeee") == "hello!"
    assert m1.hi("goorilla") == "goodbye!"

    assert m2.hi("monki") == "goodbye!"
    assert m2.hi("monkeeeee") == "goodbye!"
    assert m2.hi("goorilla") == "hello!"


def test_configure_dependent_pileon():
    m = Monkey(regex="chimpanze+") + Monkey(regex="go+rilla")
    assert m.hi("goorilla") == "hello!"



from typing import Any, Callable, Literal, Mapping, Sequence

import pytest

from ovld.core import Ovld, ovld
from ovld.types import All, Whatever


def test_literal():
    @ovld
    def f(x: Literal[0]):
        return "zero"

    @f.register
    def f(x: Literal[1]):
        return "one"

    @f.register
    def f(x: Literal[2, 3]):
        return "two or three"

    @f.register
    def f(x: object):
        return "nah"

    assert f(0) == "zero"
    assert f(1) == "one"
    assert f(2) == "two or three"
    assert f(3) == "two or three"
    assert f(4) == "nah"


def f_i(i):
    def f(x: Literal[i]):
        return x * x

    return f


def test_many_literals():
    f = Ovld()

    n = 10

    for i in range(n):
        f.register(f_i(i))

    for i in range(n):
        assert f(i) == i * i

    with pytest.raises(TypeError, match="No method"):
        f(-1)

    with pytest.raises(TypeError, match="No method"):
        f(n)


def test_tuples():
    @ovld
    def f(t: tuple[()]):
        return 0

    @ovld
    def f(t: tuple[int]):
        return 1

    @ovld
    def f(t: tuple[str]):
        return 2

    @ovld
    def f(t: tuple[int, str]):
        return 3

    @ovld
    def f(t: tuple[Literal["z"]]):
        return 4

    @ovld
    def f(t: tuple[tuple[tuple[int]]]):
        return 5

    assert f(()) == 0
    assert f((1,)) == 1
    assert f(("x",)) == 2
    assert f((2, "y")) == 3
    assert f(("z",)) == 4
    assert f((((1,),),)) == 5


def test_list():
    @ovld
    def f(li: list[int]):
        return 0

    @ovld
    def f(li: list[str]):
        return 1

    @ovld
    def f(li: Sequence[float]):
        return 2

    assert f([1, 2, 3]) == 0
    assert f(["x", "y"]) == 1
    assert f([1.5, 3.5]) == 2
    with pytest.raises(TypeError):
        f([])


def test_dict():
    @ovld
    def f(d: dict[str, int]):
        return 0

    @ovld
    def f(d: dict[int, list[int]]):
        return 1

    @ovld
    def f(d: Mapping[float, float]):
        return 1

    assert f({"x": 1, "y": 2}) == 0
    assert f({1: [0, 3], 3: [9, 0, 7], 4: []}) == 1
    with pytest.raises(TypeError):
        assert f({1: 3})
    with pytest.raises(TypeError):
        assert f({})


def test_set():
    @ovld
    def f(d: set[str]):
        return 0

    @ovld
    def f(d: set[int]):
        return 1

    assert f({"x", "y"}) == 0
    assert f({1, 7, 4}) == 1
    with pytest.raises(TypeError):
        assert f([1, 2, 3])
    with pytest.raises(TypeError):
        assert f(set())


class Animal:
    pass


class Mammal(Animal):
    pass


class Cat(Mammal):
    pass


def test_callable():
    @ovld
    def f(fn: Callable[[Mammal, Mammal], Mammal]):
        return "MM.M"

    @ovld
    def f(fn: Callable[[Cat], Animal]):
        return "C.A"

    @ovld
    def f(fn: object):
        return None

    #########

    def iscase(n):
        def deco(fn):
            assert f(fn) == n

        return deco

    #########

    @iscase("MM.M")
    def _(x: Mammal, y: Mammal) -> Mammal:
        pass

    @iscase("MM.M")
    def _(x: Animal, y: Animal) -> Cat:
        pass

    @iscase("C.A")
    def _(x: Animal) -> Cat:
        pass

    @iscase(None)
    def _(x: Mammal, y: Mammal, z) -> Mammal:
        pass

    @iscase("MM.M")
    def _(x: Mammal, y: Mammal, z=3) -> Mammal:
        pass

    @iscase(None)
    def _(x: Mammal, y: Mammal, *, z) -> Mammal:
        pass

    @iscase("MM.M")
    def _(x: Mammal, y: Mammal, *, z=3) -> Mammal:
        pass


def test_callable_whatever():
    @ovld
    def f(fn: Callable[[Whatever, Whatever], Whatever]):
        return "yes"

    @ovld
    def f(fn: object):
        return "no"

    #########

    def iscase(n):
        def deco(fn):
            assert f(fn) == n

        return deco

    #########

    @iscase("yes")
    def _(x: Mammal, y: Mammal) -> Mammal:
        pass

    @iscase("yes")
    def _(x: int, y: Cat) -> str:
        pass

    @iscase("no")
    def _(x: int) -> str:
        pass


def test_callable_all():
    @ovld
    def f(fn: Callable[[All, All], Any]):
        return "yes"

    @ovld
    def f(fn: object):
        return "no"

    #########

    def iscase(n):
        def deco(fn):
            assert f(fn) == n

        return deco

    #########

    @iscase("yes")
    def _(x: Mammal, y: Mammal) -> Mammal:
        pass

    @iscase("yes")
    def _(x: int, y: Cat) -> str:
        pass

    @iscase("no")
    def _(x: int) -> str:
        pass



"""Utilities to deal with function signatures."""

import inspect
import itertools
import typing
from collections import OrderedDict, defaultdict
from dataclasses import dataclass, field
from functools import cached_property
from types import GenericAlias

from .types import normalize_type
from .utils import MISSING, subtler_type


class LazySignature(inspect.Signature):
    def __init__(self, ovld):
        super().__init__([])
        self.ovld = ovld

    def replace(
        self, *, parameters=inspect._void, return_annotation=inspect._void
    ):  # pragma: no cover
        if parameters is inspect._void:
            parameters = self.parameters.values()

        if return_annotation is inspect._void:
            return_annotation = self._return_annotation

        return inspect.Signature(parameters, return_annotation=return_annotation)

    @property
    def parameters(self):
        anal = self.ovld.argument_analysis
        parameters = []
        if anal.is_method:
            parameters.append(
                inspect.Parameter(
                    name="self",
                    kind=inspect._POSITIONAL_ONLY,
                )
            )
        parameters += [
            inspect.Parameter(
                name=p,
                kind=inspect._POSITIONAL_ONLY,
            )
            for p in anal.strict_positional_required
        ]
        parameters += [
            inspect.Parameter(
                name=p,
                kind=inspect._POSITIONAL_ONLY,
                default=MISSING,
            )
            for p in anal.strict_positional_optional
        ]
        parameters += [
            inspect.Parameter(
                name=p,
                kind=inspect._POSITIONAL_OR_KEYWORD,
            )
            for p in anal.positional_required
        ]
        parameters += [
            inspect.Parameter(
                name=p,
                kind=inspect._POSITIONAL_OR_KEYWORD,
                default=MISSING,
            )
            for p in anal.positional_optional
        ]
        parameters += [
            inspect.Parameter(
                name=p,
                kind=inspect._KEYWORD_ONLY,
            )
            for p in anal.keyword_required
        ]
        parameters += [
            inspect.Parameter(
                name=p,
                kind=inspect._KEYWORD_ONLY,
                default=MISSING,
            )
            for p in anal.keyword_optional
        ]
        return OrderedDict({p.name: p for p in parameters})


@dataclass(frozen=True)
class Arginfo:
    position: typing.Optional[int]
    name: typing.Optional[str]
    required: bool
    ann: type

    @cached_property
    def is_complex(self):
        return isinstance(self.ann, GenericAlias)

    @cached_property
    def canonical(self):
        return self.name if self.position is None else self.position


@dataclass(frozen=True)
class Signature:
    types: tuple
    return_type: type
    req_pos: int
    max_pos: int
    req_names: frozenset
    vararg: bool
    priority: float
    tiebreak: int = 0
    is_method: bool = False
    arginfo: list[Arginfo] = field(default_factory=list, hash=False, compare=False)

    @classmethod
    def extract(cls, fn, lcl={}):
        typelist = []
        sig = inspect.signature(fn)
        max_pos = 0
        req_pos = 0
        req_names = set()
        is_method = False

        arginfo = []
        for i, (name, param) in enumerate(sig.parameters.items()):
            if name == "self" or (name == "cls" and getattr(fn, "specializer", False)):
                if i != 0:  # pragma: no cover
                    raise Exception(
                        f"Argument name '{name}' marks a method and must always be in the first position."
                    )
                is_method = True
                continue
            pos = nm = None
            ann = normalize_type(param.annotation, fn, lcl)
            if param.kind is inspect._POSITIONAL_ONLY:
                pos = i - is_method
                typelist.append(ann)
                req_pos += param.default is inspect._empty
                max_pos += 1
            elif param.kind is inspect._POSITIONAL_OR_KEYWORD:
                pos = i - is_method
                nm = param.name
                typelist.append(ann)
                req_pos += param.default is inspect._empty
                max_pos += 1
            elif param.kind is inspect._KEYWORD_ONLY:
                nm = param.name
                typelist.append((param.name, ann))
                if param.default is inspect._empty:
                    req_names.add(param.name)
            elif param.kind is inspect._VAR_POSITIONAL:
                raise TypeError("ovld does not support *args")
            elif param.kind is inspect._VAR_KEYWORD:
                raise TypeError("ovld does not support **kwargs")
            arginfo.append(
                Arginfo(
                    position=pos,
                    name=nm,
                    required=param.default is inspect._empty,
                    ann=ann,
                )
            )

        return cls(
            types=tuple(typelist),
            return_type=normalize_type(sig.return_annotation, fn),
            req_pos=req_pos,
            max_pos=max_pos,
            req_names=frozenset(req_names),
            vararg=False,
            is_method=is_method,
            priority=None,
            arginfo=arginfo,
        )


class ArgumentAnalyzer:
    def __init__(self):
        self.name_to_positions = defaultdict(set)
        self.position_to_names = defaultdict(set)
        self.counts = defaultdict(lambda: [0, 0])
        self.complex_transforms = set()
        self.total = 0
        self.is_method = None
        self.done = False

    def add(self, sig):
        self.done = False
        self.complex_transforms.update(arg.canonical for arg in sig.arginfo if arg.is_complex)
        for arg in sig.arginfo:
            if arg.position is not None:
                self.position_to_names[arg.position].add(arg.name)
            if arg.name is not None:
                self.name_to_positions[arg.name].add(arg.canonical)

            cnt = self.counts[arg.canonical]
            cnt[0] += arg.required
            cnt[1] += 1

        self.total += 1

        if self.is_method is None:
            self.is_method = sig.is_method
        elif self.is_method != sig.is_method:  # pragma: no cover
            raise TypeError(
                "Some, but not all registered methods define `self`. It should be all or none."
            )

    def compile(self):
        if self.done:
            return
        for name, pos in self.name_to_positions.items():
            if len(pos) != 1:
                if all(isinstance(p, int) for p in pos):
                    raise TypeError(
                        f"Argument '{name}' is declared in different positions by different methods. The same argument name should always be in the same position unless it is strictly positional."
                    )
                else:
                    raise TypeError(
                        f"Argument '{name}' is declared in a positional and keyword setting by different methods. It should be either."
                    )

        p_to_n = [list(names) for _, names in sorted(self.position_to_names.items())]

        positional = list(
            itertools.takewhile(
                lambda names: len(names) == 1 and isinstance(names[0], str),
                reversed(p_to_n),
            )
        )
        positional.reverse()
        strict_positional = p_to_n[: len(p_to_n) - len(positional)]

        assert strict_positional + positional == p_to_n

        self.strict_positional_required = [
            f"ARG{pos + 1}"
            for pos, _ in enumerate(strict_positional)
            if self.counts[pos][0] == self.total
        ]
        self.strict_positional_optional = [
            f"ARG{pos + 1}"
            for pos, _ in enumerate(strict_positional)
            if self.counts[pos][0] != self.total
        ]

        self.positional_required = [
            names[0]
            for pos, names in enumerate(positional)
            if self.counts[pos + len(strict_positional)][0] == self.total
        ]
        self.positional_optional = [
            names[0]
            for pos, names in enumerate(positional)
            if self.counts[pos + len(strict_positional)][0] != self.total
        ]

        keywords = [
            name for _, (name,) in self.name_to_positions.items() if not isinstance(name, int)
        ]
        self.keyword_required = [
            name for name in keywords if self.counts[name][0] == self.total
        ]
        self.keyword_optional = [
            name for name in keywords if self.counts[name][0] != self.total
        ]
        self.done = True

    def lookup_for(self, key):
        return subtler_type if key in self.complex_transforms else type



import ast
import inspect
import textwrap
from functools import reduce
from itertools import count
from types import CodeType, FunctionType

from .codegen import (
    Code,
    instantiate_code,
    rename_code,
    rename_function,
    transfer_function,
)
from .utils import MISSING, NameDatabase, SpecialForm, UsageError, is_dependent, subtler_type

recurse = SpecialForm("recurse")
call_next = SpecialForm("call_next")
resolve = SpecialForm("resolve")
current_code = SpecialForm("current_code")


dispatch_template = """
def __WRAP_DISPATCH__(OVLD):
    def __DISPATCH__({args}):
        {body}

    return __DISPATCH__
"""


call_template = """
{mvar} = OVLD.map[({lookup})]
return {mvar}({posargs})
"""


def generate_checking_code(typ):
    if hasattr(typ, "codegen"):
        return typ.codegen()
    else:
        return Code("isinstance($arg, $this)", this=typ)


def generate_dispatch(ov, arganal):
    def join(li, sep=", ", trail=False):
        li = [x for x in li if x]
        rval = sep.join(li)
        if len(li) == 1 and trail:
            rval += ","
        return rval

    arganal.compile()

    spr = arganal.strict_positional_required
    spo = arganal.strict_positional_optional
    pr = arganal.positional_required
    po = arganal.positional_optional
    kr = arganal.keyword_required
    ko = arganal.keyword_optional

    inits = set()

    kwargsstar = ""
    targsstar = ""

    args = ["self" if arganal.is_method else ""]
    body = [""]
    posargs = ["self" if arganal.is_method else ""]
    lookup = []

    i = 0
    ndb = NameDatabase(default_name="INJECT")

    def lookup_for(x):
        return ndb[arganal.lookup_for(x)]

    for name in spr + spo + pr + po + kr:
        ndb.register(name)

    mv = ndb.gensym(desired_name="method")

    for name in spr + spo:
        if name in spr:
            args.append(name)
        else:
            args.append(f"{name}=MISSING")
        posargs.append(name)
        lookup.append(f"{lookup_for(i)}({name})")
        i += 1

    if len(po) <= 1 and (spr or spo):
        # If there are more than one non-strictly positional optional arguments,
        # then all positional arguments are strictly positional, because if e.g.
        # x and y are optional we want x==MISSING to imply that y==MISSING, but
        # that only works if y cannot be provided as a keyword argument.
        args.append("/")

    for name in pr + po:
        if name in pr:
            args.append(name)
        else:
            args.append(f"{name}=MISSING")
        posargs.append(name)
        lookup.append(f"{lookup_for(i)}({name})")
        i += 1

    if len(po) > 1:
        args.append("/")

    if kr or ko:
        args.append("*")

    for name in kr:
        args.append(f"{name}")
        posargs.append(f"{name}={name}")
        lookup.append(f"({name!r}, {lookup_for(name)}({name}))")

    for name in ko:
        args.append(f"{name}=MISSING")
        kwargsstar = "**KWARGS"
        targsstar = "*TARGS"
        inits.add("KWARGS = {}")
        inits.add("TARGS = []")
        body.append(f"if {name} is not MISSING:")
        body.append(f"    KWARGS[{name!r}] = {name}")
        body.append(f"    TARGS.append(({name!r}, {lookup_for(name)}({name})))")

    posargs.append(kwargsstar)
    lookup.append(targsstar)

    fullcall = call_template.format(
        lookup=join(lookup, trail=True),
        posargs=join(posargs),
        mvar=mv,
    )

    calls = []
    if spo or po:
        req = len(spr + pr)
        for i, arg in enumerate(spo + po):
            call = call_template.format(
                lookup=join(lookup[: req + i], trail=True),
                posargs=join(posargs[: req + i + 1]),
                mvar=mv,
            )
            call = textwrap.indent(call, "        ")
            calls.append(f"\nif {arg} is MISSING:{call}")
    calls.append(fullcall)

    lines = [*inits, *body, textwrap.indent("".join(calls), "        ")]
    code = dispatch_template.format(
        args=join(args),
        body=join(lines, sep="\n        ").lstrip(),
    )
    wr = instantiate_code(
        "__WRAP_DISPATCH__", code, inject={"MISSING": MISSING, **ndb.variables}
    )
    return wr(ov)


def generate_dependent_dispatch(tup, handlers, next_call, slf, name, err, nerr):
    def to_dict(tup):
        return dict(
            entry if isinstance(entry, tuple) else (i, entry) for i, entry in enumerate(tup)
        )

    def argname(x):
        return f"ARG{x}" if isinstance(x, int) else x

    def argprovide(x):
        return f"ARG{x}" if isinstance(x, int) else f"{x}={x}"

    tup = to_dict(tup)
    handlers = [(h, to_dict(types)) for h, types in handlers]
    ndb = NameDatabase(default_name="INJECT")
    conjs = []

    exclusive = False
    keyexpr = None
    keyed = None
    for k in tup:
        featured = set(types[k] for h, types in handlers)
        if len(featured) == len(handlers):
            possibilities = set(type(t) for t in featured)
            focus = possibilities.pop()
            # Possibilities is now empty if only one type of DependentType

            if not possibilities:
                if getattr(focus, "keyable_type", False):
                    all_keys = [
                        {key: h for key in types[k].get_keys()} for h, types in handlers
                    ]
                    keyed = reduce(lambda a, b: {**a, **b}, all_keys)
                    if len(keyed) == sum(map(len, all_keys)) and len(featured) < 4:
                        exclusive = True
                        keyexpr = None
                    else:
                        keyexpr = focus.keygen().sub(arg=Code(argname(k))).fill(ndb)

                else:
                    exclusive = getattr(focus, "exclusive_type", False)

    for i, (h, types) in enumerate(handlers):
        relevant = [k for k in tup if is_dependent(types[k])]
        if len(relevant) > 1:
            # The keyexpr method only works if there is only one condition to check.
            keyexpr = keyed = None
        codes = [
            generate_checking_code(types[k]).sub(arg=Code(argname(k))).fill(ndb)
            for k in relevant
        ]
        conj = " and ".join(codes)
        if not conj:  # pragma: no cover
            # Not sure if this can happen
            conj = "True"
        conjs.append(conj)

    if len(handlers) == 1:
        exclusive = True

    argspec = ", ".join(argname(x) for x in tup)
    argcall = ", ".join(argprovide(x) for x in tup)

    body = []
    if keyexpr:
        body.append(f"HANDLER = {ndb[keyed]}.get({keyexpr}, FALLTHROUGH)")
        body.append(f"return HANDLER({slf}{argcall})")

    elif exclusive:
        for i, conj in enumerate(conjs):
            body.append(f"if {conj}: return HANDLER{i}({slf}{argcall})")
        body.append(f"return FALLTHROUGH({slf}{argcall})")

    else:
        for i, conj in enumerate(conjs):
            body.append(f"MATCH{i} = {conj}")

        summation = " + ".join(f"MATCH{i}" for i in range(len(handlers)))
        body.append(f"SUMMATION = {summation}")
        body.append("if SUMMATION == 1:")
        for i, (h, types) in enumerate(handlers):
            body.append(f"    if MATCH{i}: return HANDLER{i}({slf}{argcall})")
        body.append("elif SUMMATION == 0:")
        body.append(f"    return FALLTHROUGH({slf}{argcall})")
        body.append("else:")
        body.append(f"    raise {ndb[err]}")

    body_text = textwrap.indent("\n".join(body), "    ")
    code = f"def __DEPENDENT_DISPATCH__({slf}{argspec}):\n{body_text}"

    inject = ndb.variables
    for i, (h, types) in enumerate(handlers):
        inject[f"HANDLER{i}"] = h

    def raise_error(*args, **kwargs):
        raise nerr

    inject["FALLTHROUGH"] = (next_call and next_call[0]) or raise_error

    fn = instantiate_code(symbol="__DEPENDENT_DISPATCH__", code=code, inject=inject)
    return rename_function(fn, name)


_current = count()


class Conformer:
    __slots__ = ("code", "orig_fn", "renamed_fn", "ovld")

    def __init__(self, ovld, orig_fn, renamed_fn):
        self.ovld = ovld
        self.orig_fn = orig_fn
        self.renamed_fn = renamed_fn
        self.code = orig_fn.__code__

    def __conform__(self, new):
        if isinstance(new, FunctionType):
            new_fn = new
            new_code = new.__code__
        else:
            new_fn = None
            new_code = new

        self.ovld.unregister(self.orig_fn)

        if new_fn is None:  # pragma: no cover
            if new_code is None:
                return
            ofn = self.orig_fn
            new_fn = transfer_function(
                func=ofn,
                code=new_code,
            )

        self.ovld.register(new_fn)

        from codefind import code_registry

        code_registry.update_cache_entry(self, self.code, new_code)

        self.code = new_code


class NameConverter(ast.NodeTransformer):
    def __init__(self, anal, special_syms, mapping):
        self.analysis = anal
        self.syms = special_syms
        self.mapping = mapping
        self.ovld_mangled = mapping[recurse]
        self.map_mangled = mapping[resolve]
        self.code_mangled = mapping[current_code]
        self.count = count()

    def is_special(self, name, *kinds):
        return any(name in self.syms[kind] for kind in kinds)

    def visit_Name(self, node):
        if node.id in self.mapping:
            return ast.copy_location(
                old_node=node,
                new_node=ast.Name(self.mapping[node.id], ctx=node.ctx),
            )
        elif self.is_special(node.id, call_next):
            raise UsageError("call_next should be called right away")
        else:
            return node

    def visit_Call(self, node):
        if not isinstance(node.func, ast.Name) or not self.is_special(
            node.func.id, recurse, call_next
        ):
            return self.generic_visit(node)

        if any(isinstance(arg, ast.Starred) for arg in node.args):
            return self.generic_visit(node)

        cn = self.is_special(node.func.id, call_next)
        tmp = f"__TMP{next(self.count)}_"

        def _make_lookup_call(key, arg):
            name = (
                "__SUBTLER_TYPE" if self.analysis.lookup_for(key) is subtler_type else "type"
            )
            value = ast.NamedExpr(
                target=ast.Name(id=f"{tmp}{key}", ctx=ast.Store()),
                value=self.visit(arg),
            )
            func = ast.Name(id=name, ctx=ast.Load())
            return ast.Call(
                func=func,
                args=[value],
                keywords=[],
            )

        # type index for positional arguments
        type_parts = [_make_lookup_call(i, arg) for i, arg in enumerate(node.args)]

        # type index for keyword arguments
        type_parts += [
            ast.Tuple(
                elts=[
                    ast.Constant(value=kw.arg),
                    _make_lookup_call(kw.arg, kw.value),
                ],
                ctx=ast.Load(),
            )
            for kw in node.keywords
        ]

        if cn:
            type_parts.insert(0, ast.Name(id=self.code_mangled, ctx=ast.Load()))
        method = ast.Subscript(
            value=ast.Name(id=self.map_mangled, ctx=ast.Load()),
            slice=ast.Tuple(
                elts=type_parts,
                ctx=ast.Load(),
            ),
            ctx=ast.Load(),
        )
        if self.analysis.is_method:
            selfarg = [ast.Name(id="self", ctx=ast.Load())]
        else:
            selfarg = []

        new_node = ast.Call(
            func=method,
            args=selfarg
            + [ast.Name(id=f"{tmp}{i}", ctx=ast.Load()) for i, arg in enumerate(node.args)],
            keywords=[
                ast.keyword(
                    arg=kw.arg,
                    value=ast.Name(id=f"{tmp}{kw.arg}", ctx=ast.Load()),
                )
                for kw in node.keywords
            ],
        )
        return ast.copy_location(old_node=node, new_node=new_node)


def _search_names(co, specials, glb, closure=None):
    def _search(co, values, glb, closure=None):
        if isinstance(co, CodeType):
            if closure is not None:
                for varname, cell in zip(co.co_freevars, closure):
                    if any(cell.cell_contents is v for v in values):
                        yield varname
            for name in co.co_names:
                if any(glb.get(name, None) is v for v in values):
                    yield name
            else:
                for ct in co.co_consts:
                    yield from _search(ct, values, glb)

    return {k: list(_search(co, v, glb, closure)) for k, v in specials.items()}


def adapt_function(fn, ovld, newname):
    """Create a copy of the function with a different name."""
    syms = _search_names(
        fn.__code__,
        {
            recurse: (recurse, ovld, ovld.dispatch),
            call_next: (call_next,),
            resolve: (resolve,),
            current_code: (current_code,),
        },
        fn.__globals__,
        fn.__closure__,
    )
    if any(syms.values()):
        return recode(fn, ovld, syms, newname)
    else:
        return rename_function(fn, newname)


def closure_wrap(tree, fname, names):
    wrap = ast.copy_location(
        ast.FunctionDef(
            name="##create_closure",
            args=ast.arguments(
                posonlyargs=[],
                args=[ast.arg(arg=name) for name in names],
                vararg=None,
                kwonlyargs=[],
                kw_defaults=[],
                kwarg=None,
                defaults=[],
            ),
            body=[
                tree,
                ast.Return(ast.Name(id=fname, ctx=ast.Load())),
            ],
            decorator_list=[],
            returns=None,
        ),
        tree,
    )
    ast.fix_missing_locations(wrap)
    return ast.Module(body=[wrap], type_ignores=[])


def recode(fn, ovld, syms, newname):
    ovld_mangled = f"___OVLD{ovld.id}"
    map_mangled = f"___MAP{ovld.id}"
    code_mangled = f"___CODE{next(_current)}"
    try:
        src = inspect.getsource(fn)
    except OSError:  # pragma: no cover
        raise OSError(
            f"ovld is unable to rewrite {fn} because it cannot read its source code."
            " It may be an issue with __pycache__, so try to either change the source"
            " to force a refresh, or remove __pycache__ altogether. If that does not work,"
            " avoid calling recurse()/call_next()"
        )
    tree = ast.parse(textwrap.dedent(src))

    mapping = {
        recurse: ovld_mangled,
        resolve: map_mangled,
        current_code: code_mangled,
    }
    for special, symbols in syms.items():
        for sym in symbols:
            if special in mapping:
                mapping[sym] = mapping[special]
    new = NameConverter(
        anal=ovld.argument_analysis,
        special_syms=syms,
        mapping=mapping,
    ).visit(tree)

    new.body[0].decorator_list = []
    if fn.__closure__:
        new = closure_wrap(new.body[0], "irrelevant", fn.__code__.co_freevars)
    ast.fix_missing_locations(new)
    line_delta = fn.__code__.co_firstlineno - 1
    col_delta = len(firstline := src.split("\n", 1)[0]) - len(firstline.lstrip())
    for node in ast.walk(new):
        if hasattr(node, "lineno"):
            node.lineno += line_delta
        if hasattr(node, "end_lineno"):
            node.end_lineno += line_delta
        if hasattr(node, "col_offset"):
            node.col_offset += col_delta
        if hasattr(node, "end_col_offset"):
            node.end_col_offset += col_delta

    res = compile(new, mode="exec", filename=fn.__code__.co_filename)
    if fn.__closure__:
        res = [x for x in res.co_consts if isinstance(x, CodeType)][0]
    (*_, new_code) = [ct for ct in res.co_consts if isinstance(ct, CodeType)]
    new_closure = tuple(
        [fn.__closure__[fn.__code__.co_freevars.index(name)] for name in new_code.co_freevars]
    )
    new_fn = transfer_function(
        func=fn,
        code=rename_code(new_code, newname),
        name=newname,
        closure=new_closure,
    )
    new_fn.__globals__["__SUBTLER_TYPE"] = subtler_type
    new_fn.__globals__[ovld_mangled] = ovld.dispatch
    new_fn.__globals__[map_mangled] = ovld.map
    new_fn.__globals__[code_mangled] = new_fn.__code__
    return new_fn



from ovld import ovld, recurse

_ovld_exc = None
try:

    @ovld
    def pluralize(x: int):
        return x + 1

    @ovld
    def pluralize(x: str):
        return x + "s"

except Exception as exc:
    _ovld_exc = exc


_ovld_dispatch_exc = None
try:

    @ovld
    def roesti(x):
        return recurse(x, x)

    @ovld
    def roesti(x: int, y: int):
        return x * y

    @ovld
    def roesti(x: str, y: str):
        return x + y

    @ovld
    def roesti(xs: list, _):
        return [recurse(x) for x in xs]

except Exception as exc:
    _ovld_dispatch_exc = exc


def test_global_ovld():
    if _ovld_exc:
        raise _ovld_exc
    assert pluralize(10) == 11
    assert pluralize("alouette") == "alouettes"


def test_global_ovld_dispatch():
    if _ovld_dispatch_exc:
        raise _ovld_dispatch_exc
    assert roesti(10) == 100
    assert roesti("alouette") == "alouettealouette"



import sys
from dataclasses import dataclass
from typing import Annotated, Any, Iterable, Mapping

import pytest

from ovld.dependent import Dependent, Regexp
from ovld.mro import Order, subclasscheck, typeorder
from ovld.types import (
    All,
    Dataclass,
    HasMethod,
    Intersection,
    Order,
    Union,
    Whatever,
    typeorder,
)


def identity(x):
    return x


class A:
    pass


class B(A):
    pass


class Prox:
    _cls = object

    def __class_getitem__(cls, other):
        return type(f"Prox[{other.__name__}]", (Prox,), {"_cls": other})

    @classmethod
    def __is_supertype__(cls, other):
        return subclasscheck(other, cls._cls)

    @classmethod
    def __is_subtype__(cls, other):
        return subclasscheck(cls._cls, other)

    @classmethod
    def __type_order__(cls, other):
        return typeorder(cls._cls, other)


@dataclass
class Point:
    x: int
    y: int


def inorder(*seq):
    for a, b in zip(seq[:-1], seq[1:]):
        assert typeorder(a, b) == Order.MORE
        assert typeorder(b, a) == Order.LESS


def sameorder(*seq):
    for a, b in zip(seq[:-1], seq[1:]):
        assert typeorder(a, b) is Order.SAME
        assert typeorder(b, a) is Order.SAME


def noorder(*seq):
    for a, b in zip(seq[:-1], seq[1:]):
        assert typeorder(a, b) is Order.NONE
        assert typeorder(b, a) is Order.NONE


def test_merge():
    assert Order.merge([Order.SAME, Order.SAME]) is Order.SAME
    assert Order.merge([Order.SAME, Order.MORE]) is Order.MORE
    assert Order.merge([Order.SAME, Order.LESS]) is Order.LESS
    assert Order.merge([Order.MORE]) is Order.MORE
    assert Order.merge([Order.LESS]) is Order.LESS
    assert Order.merge([Order.LESS, Order.MORE]) is Order.NONE
    assert Order.merge([Order.LESS, Order.LESS, Order.NONE]) is Order.NONE


def test_typeorder():
    inorder(object, int)
    inorder(object, Union[int, str], str)
    inorder(object, Dataclass, Point)
    inorder(object, type, type[Dataclass])
    inorder(type[list], type[list[int]])
    inorder(str, Intersection[int, str])
    inorder(object, Intersection[object, int])
    inorder(object, Iterable, Iterable[int], list[int])
    inorder(Iterable[int], list)
    inorder(list, list[int])

    sameorder(int, int)
    sameorder(Mapping[str, int], Mapping[str, int])

    noorder(tuple[int, int], tuple[int])
    noorder(dict[str, int], dict[int, str])
    noorder(dict[str, object], dict[object, str])
    noorder(type[int], type[Dataclass])
    noorder(float, int)
    noorder(int, str)
    noorder(int, Dataclass)
    noorder(Union[int, str], float)

    inorder(Whatever, object)
    inorder(All, object)


def test_subclasscheck():
    assert subclasscheck(B, A)
    assert not subclasscheck(A, B)
    assert subclasscheck(A, A)
    assert subclasscheck(A, object)
    assert subclasscheck(A, Union[A, int])
    assert subclasscheck(int, Union[A, int])

    assert subclasscheck(All, int)
    assert not subclasscheck(int, All)

    assert subclasscheck(Whatever, int)
    assert subclasscheck(int, Whatever)


def test_subclasscheck_generic():
    assert subclasscheck(list[int], list[int])
    assert subclasscheck(list[int], list[object])
    assert not subclasscheck(list[object], list[int])


def test_subclasscheck_type():
    assert subclasscheck(type[int], type[object])


def test_subclasscheck_anntype():
    assert subclasscheck(type[Annotated[int, "hello"]], type[Annotated])
    assert subclasscheck(type[Annotated[int, "hello"]], type[Annotated[object, "hello"]])
    assert subclasscheck(
        type[Annotated[int, "hello"]], type[Annotated[object, "hello", "world"]]
    )
    assert subclasscheck(
        type[Annotated[int, "hello", "world"]], type[Annotated[object, "hello"]]
    )
    assert not subclasscheck(type[Annotated[object, "hello"]], type[Annotated[int, "hello"]])
    assert not subclasscheck(type[Annotated[int, "hello"]], type[Annotated[int, "world"]])
    assert not subclasscheck(type[Annotated[int, "hello"]], type[int])
    assert not subclasscheck(type[int], type[Annotated[int, "hello"]])


def test_subclasscheck_anntype_type():
    assert subclasscheck(type[Annotated[int, "hello"]], type[Annotated[int, str]])
    assert subclasscheck(type[Annotated[int, str]], type[Annotated[object, str]])
    assert not subclasscheck(type[Annotated[int, "hello"]], type[Annotated[int, int]])


@pytest.mark.skipif(
    sys.version_info < (3, 12), reason="Python 3.11 is more strict on Annotated"
)
def test_subclasscheck_anntype_any():
    assert subclasscheck(type[Annotated[int, "hello"]], type[Annotated[Any, "hello"]])
    assert subclasscheck(type[Annotated[123, "hello"]], type[Annotated[Any, "hello"]])


def test_typeorder_any():
    assert typeorder(int, Any) is Order.LESS
    assert typeorder(Any, int) is Order.MORE
    assert typeorder(Any, Any) is Order.SAME


def test_typeorder_exotic():
    assert typeorder(HasMethod["meth"], Annotated) is Order.NONE
    assert typeorder(Annotated, HasMethod["meth"]) is Order.NONE


@dataclass(frozen=True)
class Anno:
    name: str
    annotation_priority: int = 0


def test_typeorder_anntype():
    assert typeorder(type[Annotated], type[Annotated[int, "hello"]]) is Order.MORE
    assert typeorder(type[Annotated[int, "hello"]], type[Annotated]) is Order.LESS
    assert (
        typeorder(type[Annotated[int, "hello"]], type[Annotated[int, "world"]]) is Order.SAME
    )
    assert (
        typeorder(type[Annotated[object, "hello"]], type[Annotated[int, "world"]])
        is Order.MORE
    )
    assert typeorder(type[Annotated[int, "hello"]], type[int]) is Order.SAME
    assert typeorder(type[int], type[Annotated[int, "hello"]]) is Order.SAME

    zap = Anno("zap", 2)
    zop = Anno("zap", 5)
    assert typeorder(type[Annotated[int, zap]], type[Annotated[int, zop]]) is Order.MORE
    assert typeorder(type[Annotated[int, zop]], type[Annotated[int, zap]]) is Order.LESS


def test_subclasscheck_type_union():
    assert subclasscheck(type[Union[int, str]], type[Union])
    assert subclasscheck(type[Union], object)
    assert not subclasscheck(object, type[Union])


def test_typeorder_type_union():
    # assert typeorder(type[int | str], type[UnionType]) is Order.LESS
    assert typeorder(type[Union[int, str]], type[Union]) is Order.LESS
    assert typeorder(object, type[Union]) is Order.MORE


def test_subclasscheck_dependent():
    assert subclasscheck(B, Dependent[A, identity])
    assert not subclasscheck(Dependent[int, identity], Dependent[float, identity])


def test_typeorder_dependent():
    assert typeorder(Dependent[int, identity], float) is Order.NONE
    assert typeorder(Dependent[int, identity], Dependent[A, identity]) is Order.NONE
    assert typeorder(Dependent[int, identity], Dependent[float, identity]) is Order.NONE


def test_subclasscheck_proxy():
    assert subclasscheck(Prox[int], int)
    assert subclasscheck(int, Prox[int])
    assert subclasscheck(Prox[int], Prox)
    assert not subclasscheck(Prox, Prox[int])

    assert subclasscheck(Prox[B], A)
    assert not subclasscheck(Prox[A], B)
    assert subclasscheck(B, Prox[A])
    assert not subclasscheck(A, Prox[B])
    assert subclasscheck(Prox[B], Prox[A])

    assert typeorder(Prox[int], int) is Order.SAME
    assert typeorder(int, Prox[int]) is Order.SAME
    assert typeorder(Prox[int], Prox) is Order.LESS
    assert typeorder(Prox, Prox[int]) is Order.MORE


def test_all():
    rx = Regexp["[a-z]+"]

    aa = All[A]
    assert subclasscheck(aa, object)
    assert not subclasscheck(aa, int)
    assert subclasscheck(aa, A)
    assert subclasscheck(aa, B)
    assert not subclasscheck(aa, rx)

    ai = All[int]
    assert subclasscheck(ai, object)
    assert subclasscheck(ai, int)
    assert not subclasscheck(ai, A)
    assert not subclasscheck(ai, rx)

    ass = All[str]
    assert subclasscheck(ass, object)
    assert subclasscheck(ass, str)
    assert subclasscheck(ass, rx)  # Because rx's bound is str



import functools
import inspect
from copy import copy
from dataclasses import MISSING, dataclass, fields, make_dataclass, replace
from typing import Annotated, TypeVar, get_origin

from .core import Ovld, to_ovld
from .types import eval_annotation
from .utils import Named

ABSENT = Named("ABSENT")
CODEGEN = Named("CODEGEN")


class Combiner:
    def __init__(self, field=None):
        self.field = field

    def __set_name__(self, obj, field):
        self.field = field

    def get(self, cls):  # pragma: no cover
        raise NotImplementedError()

    def copy(self):
        return type(self)(self.field)

    def include(self, other):
        if type(self) is not type(other):
            raise TypeError("Cannot merge different combiner classes.")
        self.include_sametype(other)

    def include_sametype(self, other):  # pragma: no cover
        pass

    def juxtapose(self, impl):  # pragma: no cover
        raise NotImplementedError()


class KeepLast(Combiner):
    def __init__(self, field=None):
        super().__init__(field)
        self.impl = ABSENT

    def get(self, cls):
        return self.impl

    def include_sametype(self, other):
        self.impl = other.impl

    def juxtapose(self, impl):
        self.impl = impl


class ImplList(Combiner):
    def __init__(self, field=None, impls=None):
        super().__init__(field)
        self.impls = impls or []

    def copy(self):
        return type(self)(self.field, list(self.impls))

    def get(self, cls):
        if not self.impls:
            return ABSENT
        rval = self.wrap()
        return functools.wraps(self.impls[0])(rval)

    def wrap(self):  # pragma: no cover
        raise NotImplementedError()

    def include_sametype(self, other):
        self.impls += other.impls

    def juxtapose(self, impl):
        self.impls.append(impl)


class RunAll(ImplList):
    def wrap(_self):
        def run_all(self, *args, **kwargs):
            for impl in _self.impls:
                impl(self, *args, **kwargs)

        return run_all


class ReduceAll(ImplList):
    def wrap(_self):
        def reduce_all(self, x, *args, **kwargs):
            result = _self.impls[0](self, x, *args, **kwargs)
            for impl in _self.impls[1:]:
                result = impl(self, result, *args, **kwargs)
            return result

        return reduce_all


class ChainAll(ImplList):
    def wrap(_self):
        def chain_all(self, *args, **kwargs):
            self = _self.impls[0](self, *args, **kwargs)
            for impl in _self.impls[1:]:
                self = impl(self, *args, **kwargs)
            return self

        return chain_all


class BuildOvld(Combiner):
    def __init__(self, field=None, ovld=None):
        super().__init__(field)
        self.ovld = ovld or Ovld(name=field, linkback=True)
        self.pending = []
        if field is not None:
            self.__set_name__(None, field)

    def __set_name__(self, obj, field):
        self.ovld.rename(field)

    def get(self, cls):
        self.ovld.specialization_self = cls
        for f, arg in self.pending:
            f(arg)
        self.pending.clear()
        if self.ovld.empty():
            return ABSENT
        self.ovld.invalidate()
        return self.ovld.dispatch

    def copy(self):
        return type(self)(self.field, self.ovld.copy(linkback=True))

    def include_sametype(self, other):
        self.ovld.add_mixins(other.ovld)

    def juxtapose(self, impl):
        if ov := to_ovld(impl, force=False):
            self.pending.append((self.ovld.add_mixins, ov))
        elif inspect.isfunction(impl):
            self.pending.append((self.ovld.register, impl))
        else:  # pragma: no cover
            raise TypeError("Expected a function or ovld.")


class medley_cls_dict(dict):
    def __init__(self, bases, default_combiner=None):
        if default_combiner is None:
            (default_combiner,) = {b._ovld_default_combiner for b in bases}
        super().__init__()
        self._combiners = {}
        self._default_combiner = default_combiner
        self.set_direct("_ovld_combiners", self._combiners)
        self.set_direct("_ovld_default_combiner", default_combiner)
        self._basic = set()
        for base in bases:
            for attr, combiner in getattr(base, "_ovld_combiners", {}).items():
                if attr in self._combiners:
                    self._combiners[attr].include(combiner)
                else:
                    self._combiners[attr] = combiner.copy()

    def set_direct(self, attr, value):
        super().__setitem__(attr, value)

    def __setitem__(self, attr, value):
        if attr == "__annotations__" or attr == "__annotate_func__":
            self.set_direct(attr, value)
            return

        if attr == "__init__":
            raise Exception("Do not define __init__ in a Medley, use __post_init__.")

        if isinstance(value, Combiner):
            value.__set_name__(None, attr)
            self._combiners[attr] = value
            return

        combiner = self._combiners.get(attr, None)
        if combiner is None:
            if to_ovld(value, force=False):
                combiner = BuildOvld(attr)
            elif inspect.isfunction(value):
                combiner = self._default_combiner(attr)
            else:
                combiner = KeepLast(attr)
            self._combiners[attr] = combiner

        combiner.juxtapose(value)

    def __missing__(self, attr):
        if attr in self._combiners:
            if (value := self._combiners[attr].get(None)) is not ABSENT:
                return value
        raise KeyError(attr)


def codegen_key(*instances):
    rval = {}
    for instance in instances:
        keyd = {name: getattr(instance, name) for name in type(instance)._ovld_codegen_fields}
        rval.update(keyd)
    return rval


def specialize(cls, key):
    ns = medley_cls_dict((cls,))
    new_t = MedleyMC(cls.__name__, (cls,), ns)
    new_t._ovld_specialization_parent = cls
    for k, v in key.items():
        setattr(new_t, k, v)
    cls._ovld_codegen_fields = list(key.keys())
    return new_t


def remap_field(dc_field, require_default=False):
    if require_default:
        if dc_field.default is MISSING:
            # NOTE: we do not accept default_factory, because we need the default value to be set
            # in the class so that existing instances of classes[0] can see it.
            raise TypeError(
                f"Dataclass field '{dc_field.name}' must have a default value (not a default_factory) in order to be melded in."
            )
    dc_field = copy(dc_field)
    dc_field.kw_only = True
    return dc_field


class MedleyMC(type):
    def __subclasscheck__(cls, subclass):
        if getattr(cls, "_ovld_medleys", None):
            return all(issubclass(subclass, m) for m in cls._ovld_medleys)
        return super().__subclasscheck__(subclass)

    @classmethod
    def __prepare__(mcls, name, bases, default_combiner=None):
        return medley_cls_dict(bases, default_combiner=default_combiner)

    def __new__(mcls, name, bases, namespace, default_combiner=None):
        result = super().__new__(mcls, name, bases, namespace)
        for attr, combiner in result._ovld_combiners.items():
            if (value := combiner.get(result)) is not ABSENT:
                setattr(result, attr, value)
        dc = dataclass(result)
        dc._ovld_specialization_parent = None
        dc._ovld_specializations = {}
        dc._ovld_codegen_fields = [
            field.name
            for field in fields(dc)
            if (
                (t := eval_annotation(field.type, dc, {}, catch=True))
                and get_origin(t) is Annotated
                and CODEGEN in t.__metadata__
            )
        ]
        return dc

    def extend(cls, *others, extend_subclasses=True):
        if not others:  # pragma: no cover
            return cls
        all_fields = [(f.name, f.type, f) for f in fields(cls)]
        for other in others:
            all_fields += [(f.name, f.type, remap_field(f, True)) for f in fields(other)]
        melded = make_dataclass("_", fields=all_fields)
        for other in others:
            for k, v in vars(other).items():
                if k in ["__module__", "__firstlineno__", "__static_attributes__"]:
                    continue
                elif comb := cls._ovld_combiners.get(k):
                    comb.juxtapose(v)
                    setattr(cls, k, comb.get(cls))
                elif not k.startswith("_ovld_") and not k.startswith("__"):
                    setattr(cls, k, v)
        cls.__init__ = melded.__init__
        if extend_subclasses:
            for subcls in cls.__subclasses__():
                subothers = [o for o in others if not issubclass(subcls, o)]
                subcls.extend(*subothers, extend_subclasses=False)
        return cls

    def __add__(cls, other):
        return meld_classes((cls, other))

    def __iadd__(cls, other):
        return cls.extend(other)

    def __sub__(cls, other):
        return unmeld_classes(cls, other)

    def __call__(cls, *args, **kwargs):
        made = super().__call__(*args, **kwargs)
        if cls._ovld_codegen_fields and (keyd := codegen_key(made)):
            cls = cls._ovld_specialization_parent or cls
            key = tuple(sorted(keyd.items()))
            if key in cls._ovld_specializations:
                new_t = cls._ovld_specializations[key]
            else:
                new_t = specialize(cls, keyd)
                cls._ovld_specializations[key] = new_t
            obj = object.__new__(new_t)
            obj.__dict__.update(made.__dict__)
            return obj
        else:
            return made


def use_combiner(combiner):
    def deco(fn):
        cmb = combiner(fn.__name__)
        cmb.juxtapose(fn)
        return cmb

    return deco


class Medley(metaclass=MedleyMC, default_combiner=BuildOvld):
    __post_init__ = RunAll()
    __add__ = KeepLast()
    __sub__ = KeepLast()

    def __add__(self, other):
        if isinstance(self, type(other)) and not type(self)._ovld_codegen_fields:
            return replace(self, **vars(other))
        else:
            return meld([self, other])

    def __sub__(self, other):
        return unmeld(self, other)


def unmeld_classes(main: type, exclude: type):
    classes = tuple(c for c in main.__bases__ if c is not exclude)
    return meld_classes(classes)


_meld_classes_cache = {}


def meld_classes(classes):
    def key(cls):
        return getattr(cls, "_ovld_specialization_parent", None) or cls

    medleys = {}
    for cls in classes:
        medleys.update({key(x): x for x in getattr(cls, "_ovld_medleys", [cls])})
    for cls in classes:
        cls = key(cls)
        if not hasattr(cls, "_ovld_medleys"):
            for base in cls.mro():
                if base is not cls and base in medleys:
                    del medleys[base]

    medleys = tuple(medleys.values())
    if len(medleys) == 1:
        return medleys[0]

    cache_key = medleys
    if cache_key in _meld_classes_cache:
        return _meld_classes_cache[cache_key]

    cg_fields = set()
    dc_fields = []

    for base in medleys:
        cg_fields.update(base._ovld_codegen_fields)
        dc_fields.extend(
            (f.name, f.type, remap_field(f)) for f in base.__dataclass_fields__.values()
        )

    merged = medley_cls_dict(medleys)
    merged.set_direct("_ovld_codegen_fields", tuple(cg_fields))
    merged.set_direct("_ovld_medleys", tuple(medleys))
    merged.set_direct("__annotations__", {name: t for name, t, f in dc_fields})

    if "__qualname__" in merged._combiners:
        del merged._combiners["__qualname__"]

    result = make_dataclass(
        cls_name="+".join(sorted(c.__name__ for c in medleys)),
        bases=medleys,
        fields=dc_fields,
        kw_only=True,
        namespace=merged,
    )

    _meld_classes_cache[cache_key] = result
    return result


@functools.cache
def meld_classes_with_key(classes, key):
    key = dict(key)
    typ = meld_classes(classes)
    if not key:
        return typ
    else:
        return specialize(typ, key)


def meld(objects):
    key = codegen_key(*objects)
    classes = tuple(type(o) for o in objects)
    cls = meld_classes_with_key(classes, tuple(key.items()))
    obj = object.__new__(cls)
    for o in objects:
        for k, v in vars(o).items():
            setattr(obj, k, v)
    return obj


def unmeld(obj: object, exclude: type):
    if type(obj)._ovld_codegen_fields:  # pragma: no cover
        raise TypeError("Cannot unmeld an object with codegen fields")
    cls = unmeld_classes(type(obj), exclude)
    values = {}
    excluded = exclude.__dataclass_fields__
    for f in cls.__dataclass_fields__.values():
        if f.name not in excluded:
            values[f.name] = getattr(obj, f.name)
    return cls(**values)


T = TypeVar("T")
CodegenParameter = Annotated[T, CODEGEN]



from typing import TYPE_CHECKING

from . import abc  # noqa: F401
from .codegen import Code, Def, Lambda, code_generator
from .core import (
    Ovld,
    OvldBase,
    OvldMC,
    extend_super,
    is_ovld,
    ovld,
)
from .dependent import (
    Dependent,
    DependentType,
    ParametrizedDependentType,
    dependent_check,
)
from .medley import (
    CodegenParameter,
    Medley,
)
from .mro import (
    TypeRelationship,
    subclasscheck,
    typeorder,
)
from .recode import (
    call_next,
    current_code,
    recurse,
    resolve,
)
from .typemap import (
    MultiTypeMap,
    TypeMap,
)
from .types import (
    Dataclass,
    Deferred,
    Exactly,
    Intersection,
    StrictSubclass,
    class_check,
    parametrized_class_check,
)
from .utils import (
    BOOTSTRAP,
    MISSING,
    CodegenInProgress,
    Named,
    NameDatabase,
    keyword_decorator,
)
from .version import version as __version__

if TYPE_CHECKING:  # pragma: no cover
    # Pretend that @ovld is @typing.overload.
    # I can't believe this works.
    from typing import overload as ovld


__all__ = [
    "MultiTypeMap",
    "Ovld",
    "OvldBase",
    "OvldMC",
    "TypeMap",
    "extend_super",
    "is_ovld",
    "ovld",
    "Code",
    "Def",
    "Lambda",
    "code_generator",
    "Dependent",
    "ParametrizedDependentType",
    "DependentType",
    "dependent_check",
    "Medley",
    "CodegenParameter",
    "BOOTSTRAP",
    "CodegenInProgress",
    "MISSING",
    "Dataclass",
    "Named",
    "NameDatabase",
    "Deferred",
    "Exactly",
    "Intersection",
    "StrictSubclass",
    "class_check",
    "subclasscheck",
    "typeorder",
    "TypeRelationship",
    "parametrized_class_check",
    "keyword_decorator",
    "call_next",
    "recurse",
    "resolve",
    "current_code",
    "__version__",
]



"""Utilities to overload functions for multiple types."""

import inspect
import itertools
import sys
import textwrap
from dataclasses import replace
from functools import partial
from types import FunctionType

from .recode import (
    Conformer,
    adapt_function,
    generate_dispatch,
    rename_code,
)
from .signatures import ArgumentAnalyzer, LazySignature, Signature
from .typemap import MultiTypeMap
from .utils import (
    MISSING,
    ResolutionError,
    UsageError,
    keyword_decorator,
    sigstring,
    subtler_type,
)

_orig_getdoc = inspect.getdoc


def _getdoc(fn):
    if hasattr(fn, "__calculate_doc__"):
        if inspect.ismethod(fn):
            fn = fn.__func__
        fn.__doc__ = fn.__calculate_doc__()
        del fn.__calculate_doc__
    return _orig_getdoc(fn)


inspect.getdoc = _getdoc


_current_id = itertools.count()


def bootstrap_dispatch(ov, name):
    def first_entry(*args, **kwargs):
        ov.compile()
        return ov.dispatch(*args, **kwargs)

    dispatch = FunctionType(
        rename_code(first_entry.__code__, name),
        {},
        name,
        (),
        first_entry.__closure__,
    )
    dispatch.__signature__ = LazySignature(ov)
    dispatch.__ovld__ = ov
    dispatch.register = ov.register
    dispatch.resolve_for_values = ov.resolve_for_values
    dispatch.resolve = ov.resolve
    dispatch.resolve_all = ov.resolve_all
    dispatch.copy = ov.copy
    dispatch.variant = ov.variant
    dispatch.display_methods = ov.display_methods
    dispatch.display_resolution = ov.display_resolution
    dispatch.add_mixins = ov.add_mixins
    dispatch.unregister = ov.unregister
    dispatch.next = ov.next
    dispatch.first_entry = first_entry
    return dispatch


class Ovld:
    """Overloaded function.

    A function can be added with the `register` method. One of its parameters
    should be annotated with a type, but only one, and every registered
    function should annotate the same parameter.

    Arguments:
        mixins: A list of Ovld instances that contribute functions to this
            Ovld.
        name: Optional name for the Ovld. If not provided, it will be
            gotten automatically from the first registered function or
            dispatch.
        linkback: Whether to keep a pointer in the parent mixins to this
            ovld so that updates can be propagated. (default: False)
    """

    def __init__(
        self,
        *,
        mixins=[],
        name=None,
        linkback=False,
    ):
        """Initialize an Ovld."""
        self.id = next(_current_id)
        self.specialization_self = MISSING
        self._compiled = False
        self._signatures = None
        self._argument_analysis = None
        self.linkback = linkback
        self.children = []
        self.name = name
        self.shortname = name or f"__OVLD{self.id}"
        self.__name__ = name
        self._regs = {}
        self._locked = False
        self.mixins = []
        self.dispatch = bootstrap_dispatch(self, name=self.shortname)
        self.add_mixins(*mixins)

    def regs(self):
        for mixin in self.mixins:
            yield from mixin.regs()
        yield from self._regs.items()

    def empty(self):
        return not self._regs and all(m.empty() for m in self.mixins)

    def mkdoc(self):
        fns = [f for f, _ in self.regs()]
        try:
            docs = [fn.__doc__ for fn in fns if fn.__doc__]
            if len(docs) == 1:
                maindoc = docs[0]
            else:
                maindoc = f"Ovld with {len(fns)} methods."

            doc = f"{maindoc}\n\n"
            for fn in fns:
                fndef = inspect.signature(fn)
                fdoc = fn.__doc__
                if not fdoc or fdoc == maindoc:
                    doc += f"{self.__name__}{fndef}\n\n"
                else:
                    if not fdoc.strip(" ").endswith("\n"):
                        fdoc += "\n"
                    fdoc = textwrap.indent(fdoc, " " * 4)
                    doc += f"{self.__name__}{fndef}\n{fdoc}\n"
        except Exception as exc:  # pragma: no cover
            doc = f"An exception occurred when calculating the docstring: {exc}"
        return doc

    @property
    def __doc__(self):
        self.ensure_compiled()
        return self.mkdoc()

    @property
    def __signature__(self):
        return self.dispatch.__signature__

    def lock(self):
        self._locked = True

    def _attempt_modify(self):
        if self._locked:
            raise Exception(f"ovld {self} is locked for modifications")

    def add_mixins(self, *mixins):
        self._attempt_modify()
        mixins = [o for m in mixins if (o := to_ovld(m)) is not self]
        for mixin in mixins:
            if self.linkback:
                mixin.children.append(self)
        self.mixins += mixins

    def _key_error(self, key, possibilities=None):
        typenames = sigstring(key)
        if not possibilities:
            return ResolutionError(f"No method in {self} for argument types [{typenames}]")
        else:
            hlp = ""
            for c in possibilities:
                hn = getattr(c.handler, "__orig_name__", c.handler.__name__)
                hlp += (
                    f"* {hn}  (priority: {c.priority}, specificity: {list(c.specificity)})\n"
                )
            return ResolutionError(
                f"Ambiguous resolution in {self} for"
                f" argument types [{typenames}]\n"
                f"Candidates are:\n{hlp}"
                "Note: you can use @ovld(priority=X) to give higher priority to an overload."
            )

    def rename(self, name, shortname=None):
        """Rename this Ovld."""
        if name != self.name:
            self.name = name
            self.shortname = shortname or name
            self.__name__ = self.shortname
            self.dispatch = bootstrap_dispatch(self, name=self.shortname)

    def __set_name__(self, inst, name):
        self.rename(name)

    def ensure_compiled(self):
        if not self._compiled:
            self.compile()

    @property
    def signatures(self):
        if self._signatures is None:
            regs = {}
            for fn, priority in self.regs():
                ss = self.specialization_self
                cgf = getattr(ss, "_ovld_codegen_fields", ())
                lcl = {f: getattr(ss, f) for f in cgf}
                sig = replace(Signature.extract(fn, lcl), priority=priority)

                def _set(sig, fn):
                    if sig in regs:
                        # Push down the existing handler with a lower tiebreak
                        msig = replace(sig, tiebreak=sig.tiebreak - 1)
                        _set(msig, regs[sig])
                    regs[sig] = fn

                _set(sig, fn)
            self._signatures = regs
        return self._signatures

    @property
    def argument_analysis(self):
        if self._argument_analysis is None:
            aa = ArgumentAnalyzer()
            for sig in self.signatures:
                aa.add(sig)
            aa.compile()
            self._argument_analysis = aa
        return self._argument_analysis

    def compile(self):
        """Finalize this overload.

        This will populate the type maps and replace the functions decorated
        with _compile_first (__call__, __get__, etc.) with versions that assume
        the ovld has been compiled.

        This will also lock this ovld's parent mixins to prevent their
        modification.
        """
        for mixin in self.mixins:
            if self not in mixin.children:
                mixin.lock()

        if self.name is None:
            self.name = self.__name__ = f"ovld{self.id}"

        name = self.__name__
        self.map = MultiTypeMap(name=name, key_error=self._key_error, ovld=self)

        for sig, fn in list(self.signatures.items()):
            self.register_signature(sig, fn)

        dispatch = generate_dispatch(self, self.argument_analysis)
        self.dispatch.__code__ = rename_code(dispatch.__code__, self.shortname)
        self.dispatch.__kwdefaults__ = dispatch.__kwdefaults__
        self.dispatch.__annotations__ = dispatch.__annotations__
        self.dispatch.__defaults__ = dispatch.__defaults__
        self.dispatch.__globals__.update(dispatch.__globals__)
        self.dispatch.map = self.map
        self.dispatch.__generate_doc__ = self.mkdoc

        self._compiled = True

    def resolve_for_values(self, *args):
        """Find the correct method to call for the given arguments."""
        self.ensure_compiled()
        return self.map[tuple(map(subtler_type, args))]

    def resolve(self, *args, after=None):
        """Find the correct method to call for the given argument types."""
        self.ensure_compiled()
        if after:
            return self.map[(getattr(after, "__code__", after), *args)]
        else:
            return self.map[args]

    def resolve_all(self, *args, **kwargs):
        """Yield all methods that match the arguments, in priority order."""
        self.ensure_compiled()
        return self.map.resolve_all(*args, **kwargs)

    def register_signature(self, sig, orig_fn):
        """Register a function for the given signature."""
        fn = adapt_function(orig_fn, self, f"{self.__name__}[{sigstring(sig.types)}]")
        # We just need to keep the Conformer pointer alive for jurigged
        # to find it, if jurigged is used with ovld
        fn._conformer = Conformer(self, orig_fn, fn)
        self.map.register(sig, fn)
        return self

    def register(self, fn=None, priority=0):
        """Register a function."""
        if fn is None:
            return partial(self._register, priority=priority)
        priority = getattr(fn, "priority", priority)
        return self._register(fn, priority)

    def _register(self, fn, priority):
        """Register a function."""

        if not isinstance(priority, tuple):
            priority = (priority,)

        self._attempt_modify()
        if self.name is None:
            self.__qualname__ = fn.__qualname__
            self.__module__ = fn.__module__
            self.rename(f"{fn.__module__}.{fn.__qualname__}", fn.__name__)
        self._regs[fn] = priority

        self.invalidate()
        return self

    def unregister(self, fn):
        """Unregister a function."""
        self._attempt_modify()
        del self._regs[fn]
        self.invalidate()

    def invalidate(self):
        self._signatures = None
        self._argument_analysis = None
        if self._compiled:
            self._compiled = False
            self.dispatch.__code__ = self.dispatch.first_entry.__code__
        for child in self.children:
            child.invalidate()
        if hasattr(self, "dispatch"):
            self.dispatch.__calculate_doc__ = self.mkdoc

    def copy(self, mixins=[], linkback=False):
        """Create a copy of this Ovld.

        New functions can be registered to the copy without affecting the
        original.
        """
        return Ovld(mixins=[self, *mixins], linkback=linkback)

    def variant(self, fn=None, priority=0, **kwargs):
        """Decorator to create a variant of this Ovld.

        New functions can be registered to the variant without affecting the
        original.
        """
        ov = self.copy(**kwargs)
        if fn is None:
            return partial(ov.register, priority=priority)
        else:
            ov.register(fn, priority=priority)
            return ov

    def __get__(self, obj, cls):
        return self.dispatch.__get__(obj, cls)

    def __call__(self, *args, **kwargs):  # pragma: no cover
        """Call the overloaded function.

        This should be replaced by an auto-generated function.
        """
        return self.dispatch(*args, **kwargs)

    def next(self, *args):
        """Call the next matching method after the caller, in terms of priority or specificity."""
        fr = sys._getframe(1)
        key = (fr.f_code, *map(subtler_type, args))
        method = self.map[key]
        return method(*args)

    def __repr__(self):
        return f"<Ovld {self.name or hex(id(self))}>"

    def display_methods(self):
        self.ensure_compiled()
        self.map.display_methods()

    def display_resolution(self, *args, **kwargs):
        self.ensure_compiled()
        self.map.display_resolution(*args, **kwargs)


def is_ovld(x):
    """Return whether the argument is an ovld function/method."""
    return isinstance(x, Ovld) or isinstance(getattr(x, "__ovld__", False), Ovld)


def to_ovld(x, force=True):
    """Return the argument as an Ovld."""
    x = getattr(x, "__ovld__", x)
    if inspect.isfunction(x):
        return (ovld(x, fresh=True).__ovld__) if force else None
    else:
        return x if isinstance(x, Ovld) else None


def extend_super(fn):
    """Declare that this method extends the super method with more types.

    This produces an ovld using the superclass method of the same name,
    plus this definition and others with the same name.
    """
    if not is_ovld(fn):
        fn = ovld(fn, fresh=True)
    fn._extend_super = True
    return fn


class ovld_cls_dict(dict):
    """A dict for use with OvldMC.__prepare__.

    Setting a key that already corresponds to an Olvd extends that Ovld.
    """

    def __init__(self, bases):
        self._bases = bases

    def __setitem__(self, attr, value):
        prev = None
        if attr in self:
            prev = to_ovld(self[attr])
        elif is_ovld(value) and getattr(value, "_extend_super", False):
            mixins = []
            for base in self._bases:
                if (candidate := getattr(base, attr, None)) is not None:
                    if mixin := to_ovld(candidate):
                        mixins.append(mixin)
            if mixins:
                prev, *others = mixins
                prev = prev.copy()
                for other in others:
                    prev.add_mixins(other)
        else:
            prev = None

        if prev is not None:
            if is_ovld(value) and prev is not value:
                value = getattr(value, "__ovld__", value)
                if prev.name is None:
                    prev.rename(value.name)
                prev.add_mixins(value)
                value = prev
            elif inspect.isfunction(value):
                prev.register(value)
                value = prev

        super().__setitem__(attr, value.dispatch if isinstance(value, Ovld) else value)


class OvldMC(type):
    """Metaclass that allows overloading.

    A class which uses this metaclass can define multiple functions with
    the same name and different type signatures.
    """

    def create_subclass(cls, *bases, name=None):
        """Create a new subclass with the given extra bases."""
        name = name or "UnnamedOvld"
        bases = (cls, *bases)
        return type(cls)(name, bases, cls.__prepare__(name, bases))

    @classmethod
    def __prepare__(metacls, name, bases):
        d = ovld_cls_dict(bases)

        names = set()
        for base in bases:
            names.update(dir(base))

        for name in names:
            values = [getattr(base, name, None) for base in bases]
            ovlds = [v for v in values if is_ovld(v)]
            mixins = [v for v in ovlds[1:] if getattr(v, "_extend_super", False)]
            if mixins:
                o = ovlds[0].copy(mixins=mixins)
                others = [v for v in values if v is not None and not is_ovld(v)]
                for other in others:
                    o.register(other)
                o.rename(name)
                d[name] = o

        return d

    def __init__(cls, name, bases, d):
        for val in d.values():
            if o := to_ovld(val, force=False):
                o.specialization_self = cls
        super().__init__(name, bases, d)


class OvldBase(metaclass=OvldMC):
    """Base class that allows overloading of methods."""


def _find_overload(fn, **kwargs):
    fr = sys._getframe(1)  # We typically expect to get to frame 3.
    while fr and fn.__code__ not in fr.f_code.co_consts:
        # We are basically searching for the function's code object in the stack.
        # When a class/function A is nested in a class/function B, the former's
        # code object is in the latter's co_consts. If ovld is used as a decorator,
        # on A, then necessarily we are inside the execution of B, so B should be
        # on the stack and we should be able to find A's code object in there.
        fr = fr.f_back

    if not fr:
        raise UsageError("@ovld only works as a decorator.")

    dispatch = fr.f_locals.get(fn.__name__, None)

    if dispatch is None:
        dispatch = Ovld(**kwargs)
    elif not is_ovld(dispatch):  # pragma: no cover
        raise TypeError("@ovld requires Ovld instance")
    elif kwargs:  # pragma: no cover
        raise TypeError("Cannot configure an overload that already exists")
    return getattr(dispatch, "__ovld__", dispatch)


@keyword_decorator
def ovld(fn, priority=0, fresh=False, **kwargs):
    """Overload a function.

    Overloading is based on the function name.

    The decorated function should have one parameter annotated with a type.
    Any parameter can be annotated, but only one, and every overloading of a
    function should annotate the same parameter.

    The decorator optionally takes keyword arguments, *only* on the first
    use.

    Arguments:
        fn: The function to register.
        priority: The priority of the function in the resolution order.
        fresh: Whether to create a new ovld or try to reuse an existing one.
        mixins: A list of Ovld instances that contribute functions to this
            Ovld.
        name: Optional name for the Ovld. If not provided, it will be
            gotten automatically from the first registered function or
            dispatch.
        linkback: Whether to keep a pointer in the parent mixins to this
            ovld so that updates can be propagated. (default: False)
    """
    if fresh:
        dispatch = Ovld(**kwargs)
    else:
        dispatch = _find_overload(fn, **kwargs)
    dispatch.register(fn, priority=priority)
    return dispatch.dispatch



import typing
from collections.abc import Callable, Collection, Mapping, Sequence

from .dependent import Callable as OvldCallable
from .dependent import (
    CollectionFastCheck,
    Equals,
    MappingFastCheck,
    ProductType,
    SequenceFastCheck,
)
from .types import normalize_type


@normalize_type.register_generic(typing.Literal)
def _(self, t, fn):
    return Equals[t.__args__]


@normalize_type.register_generic(tuple)
def _(self, t, fn):
    args = tuple(self(arg, fn) for arg in t.__args__)
    return ProductType[args]


@normalize_type.register_generic(Sequence)
def _(self, t, fn):
    args = tuple(self(arg, fn) for arg in t.__args__)
    return SequenceFastCheck[args]


@normalize_type.register_generic(Collection)
def _(self, t, fn):
    args = tuple(self(arg, fn) for arg in t.__args__)
    return CollectionFastCheck[args]


@normalize_type.register_generic(Mapping)
def _(self, t, fn):
    args = tuple(self(arg, fn) for arg in t.__args__)
    return MappingFastCheck[args]


@normalize_type.register_generic(Callable)
def _(self, t, fn):
    *at, rt = t.__args__
    at = tuple(self(arg, fn) for arg in at)
    return OvldCallable[at, self(rt, fn)]



from dataclasses import dataclass
from enum import Enum
from graphlib import TopologicalSorter
from typing import Annotated, Any, get_args, get_origin

from .utils import UnionTypes, is_dependent


class Order(Enum):
    LESS = -1
    MORE = 1
    SAME = 0
    NONE = None

    def opposite(self):
        if self is Order.LESS:
            return Order.MORE
        elif self is Order.MORE:
            return Order.LESS
        else:
            return self

    @staticmethod
    def merge(orders):
        orders = set(orders)
        if orders == {Order.SAME}:
            return Order.SAME
        elif not (orders - {Order.LESS, Order.SAME}):
            return Order.LESS
        elif not (orders - {Order.MORE, Order.SAME}):
            return Order.MORE
        else:
            return Order.NONE


@dataclass
class TypeRelationship:
    order: Order
    supertype: bool = NotImplemented
    subtype: bool = NotImplemented


def typeorder(t1, t2):
    """Order relation between two types.

    Returns a member of the Order enum.

    * typeorder(t1, t2) is Order.LESS   if t2 is more general than t1
    * typeorder(t1, t2) is Order.SAME   if t1 is equal to t2
    * typeorder(t1, t2) is Order.MORE   if t1 is more general than t2
    * typeorder(t1, t2) is Order.NONE   if they cannot be compared
    """
    if t1 == t2:
        return Order.SAME
    if t1 is Any:
        return Order.MORE
    if t2 is Any:
        return Order.LESS

    if (
        hasattr(t1, "__type_order__")
        and (result := t1.__type_order__(t2)) is not NotImplemented
    ):
        return result
    elif (
        hasattr(t2, "__type_order__")
        and (result := t2.__type_order__(t1)) is not NotImplemented
    ):
        return result.opposite()

    o1 = get_origin(t1)
    o2 = get_origin(t2)

    if o1 is Annotated and o2 is Annotated:
        t1, *a1 = get_args(t1)
        t2, *a2 = get_args(t2)
        p1 = max([getattr(ann, "annotation_priority", 0) for ann in a1], default=0)
        p2 = max([getattr(ann, "annotation_priority", 0) for ann in a2], default=0)
        if p1 < p2:
            return Order.MORE
        elif p2 < p1:
            return Order.LESS
        else:
            return typeorder(t1, t2)

    if o1 is Annotated:
        if t2 is Annotated:
            return Order.LESS
        return typeorder(get_args(t1)[0], t2)
    if o2 is Annotated:
        if t1 is Annotated:
            return Order.MORE
        return typeorder(t1, get_args(t2)[0])

    if o2 and not o1:
        return typeorder(t2, t1).opposite()

    if o1:
        if not o2:
            order = typeorder(o1, t2)
            if order is Order.SAME:
                order = Order.LESS
            return order

        if (order := typeorder(o1, o2)) is not Order.SAME:
            return order

        args1 = get_args(t1)
        args2 = get_args(t2)

        if args1 and not args2:
            return Order.LESS
        if args2 and not args1:
            return Order.MORE
        if len(args1) != len(args2):
            return Order.NONE

        ords = [typeorder(a1, a2) for a1, a2 in zip(args1, args2)]
        return Order.merge(ords)

    if not isinstance(t1, type) or not isinstance(t2, type):
        return Order.NONE

    sx = issubclass(t1, t2)
    sy = issubclass(t2, t1)
    if sx and sy:  # pragma: no cover
        # Not sure when t1 != t2 and that happens
        return Order.SAME
    elif sx:
        return Order.LESS
    elif sy:
        return Order.MORE
    else:
        return Order.NONE


def _find_ann(main, others):
    if main in others:
        return True
    elif isinstance(main, type):
        return any(isinstance(x, main) for x in others)
    else:
        return False


def subclasscheck(t1, t2):
    """Check whether t1 is a "subclass" of t2."""
    if t1 == t2 or t2 is Any:
        return True

    if (
        hasattr(t2, "__is_supertype__")
        and (result := t2.__is_supertype__(t1)) is not NotImplemented
    ):
        return result

    if (
        hasattr(t1, "__is_subtype__")
        and (result := t1.__is_subtype__(t2)) is not NotImplemented
    ):
        return result

    if is_dependent(t2):
        # t2's instancecheck could return anything, and unless it defines
        # __is_supertype__ or __is_subtype__ the bound devolves to object
        return True

    if t2 in UnionTypes:
        return isinstance(t1, t2)

    o1 = get_origin(t1)
    o2 = get_origin(t2)

    if o1 is Annotated and o2 is Annotated:
        t1, *a1 = get_args(t1)
        t2, *a2 = get_args(t2)
        return subclasscheck(t1, t2) and any(_find_ann(main, a1) for main in a2)

    if o1 is Annotated:
        return t2 is Annotated

    if not isinstance(o1, type):
        o1 = None
    if not isinstance(o2, type):
        o2 = None

    if (o1 or o2) and o2 not in UnionTypes:
        o1 = o1 or t1
        o2 = o2 or t2
        if isinstance(o1, type) and isinstance(o2, type) and issubclass(o1, o2):
            if o2 is t2:  # pragma: no cover
                return True
            else:
                args1 = get_args(t1)
                args2 = get_args(t2)
                if len(args1) != len(args2):
                    return False
                return all(subclasscheck(a1, a2) for a1, a2 in zip(args1, args2))
        else:
            return False
    else:
        try:
            return issubclass(t1, t2)
        except TypeError:
            return False


def sort_types(cls, avail):
    # We filter everything except subclasses and dependent types that *might* cover
    # the object represented by cls.
    avail = [t for t in avail if subclasscheck(cls, t)]
    deps = {t: set() for t in avail}
    for i, t1 in enumerate(avail):
        for t2 in avail[i + 1 :]:
            # NOTE: this is going to scale poorly when there's a hundred Literal in the pool
            order = typeorder(t1, t2)
            if order is Order.LESS:
                deps[t2].add(t1)
            elif order is Order.MORE:
                deps[t1].add(t2)
    sorter = TopologicalSorter(deps)
    sorter.prepare()
    while sorter.is_active():
        nodes = sorter.get_ready()
        yield nodes
        for n in nodes:
            sorter.done(n)
